@@ -63,6 +63,18 @@ isNullPointer <- function(pointer) {
   .Call("isNullPointer", pointer, PACKAGE="TMB")
 }
 
+## Add external pointer finalizer
+registerFinalizer <- function(ADFun, DLL) {
+    finalizer <- function(ptr) {
+        if ( ! isNullPointer(ptr) ) {
+            .Call("FreeADFunObject", ptr, PACKAGE=DLL)
+        } else {
+            ## Nothing to free
+        }
+    }
+    reg.finalizer(ADFun$ptr, finalizer)
+}
+
 ##' Construct objective functions with derivatives based on the users C++ template.
 ##'
 ##' A call to \code{MakeADFun} will return an object that, based on the users DLL code (specified through \code{DLL}), contains functions to calculate the objective function
@@ -137,7 +149,7 @@ isNullPointer <- function(pointer) {
 ##' @note Do not rely upon the default arguments of any of the functions in the model object \code{obj$fn}, \code{obj$gr}, \code{obj$he}, \code{obj$report}. I.e. always use the explicit form \code{obj$fn(obj$par)} rather than \code{obj$fn()}.
 ##'
 ##' @title Construct objective functions with derivatives based on a compiled C++ template.
-##' @param data List of data objects (vectors,matrices,arrays,factors,sparse matrices) required by the user template (order does not matter and un-used components are allowed).
+##' @param data List of data objects (vectors, matrices, arrays, factors, sparse matrices) required by the user template (order does not matter and un-used components are allowed).
 ##' @param parameters List of all parameter objects required by the user template (both random and fixed effects).
 ##' @param map List defining how to optionally collect and fix parameters - see details.
 ##' @param type Character vector defining which operation stacks are generated from the users template - see details.
@@ -351,6 +363,7 @@ MakeADFun <- function(data, parameters, map=list(),
       ## User template contains atomic functions ==>
       ## Have to call "double-template" to trigger tape generation
       Fun <<- .Call("MakeDoubleFunObject",data,parameters,reportenv,PACKAGE=DLL)
+      registerFinalizer(Fun, DLL)
       ## Hack: unlist(parameters) only guarantied to be a permutation of the parameter vecter.
       out <- .Call("EvalDoubleFunObject", Fun$ptr, unlist(parameters),
                    control = list(do_simulate = as.integer(0),
@@ -404,6 +417,8 @@ MakeADFun <- function(data, parameters, map=list(),
     if("ADFun"%in%type){
       ADFun <<- .Call("MakeADFunObject",data,parameters,reportenv,
                      control=list(report=as.integer(ADreport)),PACKAGE=DLL)
+      if (!is.null(ADFun)) ## ADFun=NULL used by sdreport
+          registerFinalizer(ADFun, DLL)
       if (set.defaults) {
           par <<- attr(ADFun$ptr,"par")
           last.par <<- par
@@ -413,10 +428,14 @@ MakeADFun <- function(data, parameters, map=list(),
           value.best <<- Inf
       }
     }
-    if("Fun"%in%type)
+    if("Fun"%in%type) {
       Fun <<- .Call("MakeDoubleFunObject",data,parameters,reportenv,PACKAGE=DLL)
-    if("ADGrad"%in%type)
+      registerFinalizer(Fun, DLL)
+    }
+    if("ADGrad"%in%type) {
       ADGrad <<- .Call("MakeADGradObject",data,parameters,reportenv,PACKAGE=DLL)
+      registerFinalizer(ADGrad, DLL)
+    }
     ## Skip fixed effects from the full hessian ?
     ## * Probably more efficient - especially in terms of memory.
     ## * Only possible if a taped gradient is available - see function "ff" below.
@@ -840,8 +859,10 @@ MakeADFun <- function(data, parameters, map=list(),
            ## If no atomics on tape we have all orders implemented:
            if(!atomic) return( f(x,order=2) )
            ## Otherwise, get Hessian as 1st order derivative of gradient:
-           if(is.null(ADGrad))
+           if(is.null(ADGrad)) {
              ADGrad <<- .Call("MakeADGradObject",data,parameters,reportenv,PACKAGE=DLL)
+             registerFinalizer(ADGrad, DLL)
+           }
            f(x,type="ADGrad",order=1)
          },
          hessian=hessian, method=method,
@@ -890,6 +911,58 @@ MakeADFun <- function(data, parameters, map=list(),
   }
 }## end{ MakeADFun }
 
+##' Free memory allocated on the C++ side by \code{MakeADFun}.
+##'
+##' @note
+##' This function is normally not needed.
+##' @details
+##' An object returned by \code{MakeADFun} contains pointers to
+##' structures allocated on the C++ side. These are managed by R's
+##' garbage collector which for the most cases is sufficient. However,
+##' because the garbage collector is unware of the C++ object sizes,
+##' it may fail to release memory to the system as frequently as
+##' necessary. In such cases one can manually call
+##' \code{FreeADFun(obj)} to release the resources.
+##' @section Memory management:
+##' Memory allocated on the C++ side by \code{MakeADFun} is
+##' represented by external pointers. Each such pointer has an
+##' associated 'finalizer' (see \code{reg.finalizer}) that deallocates
+##' the external pointer when \code{gc()} decides the pointer is no
+##' longer needed.  Deallocated pointers are recognized on the R
+##' side as external null pointers \code{<pointer: (nil)>}. This is
+##' important as it provides a way to prevent the finalizers from
+##' freeing pointers that have already been deallocated \emph{even if
+##' the deallocation C-code has been unloaded}.
+##' The user DLL maintains a list of all external pointers on the C
+##' side. Three events can reduce the list:
+##' \itemize{
+##'   \item Garbage collection of an external pointer that is no longer needed (triggers corresponding finalizer).
+##'   \item Explicit deallocation of external pointers using \code{FreeADFun()} (corresponding finalizers are untriggered but harmless).
+##'   \item Unload/reload of the user's DLL deallocates all external pointers (corresponding finalizers are untriggered but harmless).
+##' }
+##' @title Free memory allocated on the C++ side by \code{MakeADFun}.
+##' @param obj Object returned by \code{MakeADFun}
+##' @return NULL
+##' @examples
+##' runExample("simple", thisR = TRUE)          ## Create 'obj'
+##' FreeADFun(obj)                              ## Free external pointers
+##' obj$fn()                                    ## Re-allocate external pointers
+FreeADFun <- function(obj) {
+    free <- function(ADFun) {
+        if (! is.null(ADFun) ) {
+            if ( ! isNullPointer(ADFun$ptr) ) {
+                .Call("FreeADFunObject", ADFun$ptr, PACKAGE = obj$env$DLL)
+            }
+        }
+    }
+    free(obj$env$Fun)
+    free(obj$env$ADFun)
+    free(obj$env$ADGrad)
+    ADHess <- environment(obj$env$spHess)$ADHess
+    free(ADHess)
+    return(NULL)
+}
+
 .removeComments <- function(x){
   x <- paste(x,collapse="\n")
   remlong <- function(x)gsub("/\\*.*?\\*/","",x)
@@ -1443,6 +1516,7 @@ sparseHessianFun <- function(obj, skipFixedEffects=FALSE) {
                   obj$env$reportenv,
                   skip, ## <-- Skip this index vector of parameters
                   PACKAGE=obj$env$DLL)
+  registerFinalizer(ADHess, obj$env$DLL)
   ev <- function(par)
           .Call("EvalADFunObject", ADHess$ptr, par,
                 control = list(
